import copy
import json
import os

import jsonschema

CONFIG_DEFAULT_PORT = 443
CONFIG_DEFAULT_CACHE_PARAMS = {
    'filename': '/tmp/cached-vms.json',
    'expiration_seconds': 86400
}

CONFIG_SCHEMA = {
    "$schema": "http://json-schema.org/draft-07/schema#",

    "definitions": {
        "cache": {
            "type": "object",
            "properties": {
                "filename": {"type": "string"},
                "expiration_seconds": {
                    "type": "number",
                    "minimum": 0
                },
            },
            "additionalProperties": False
        },
        "datacenter": {
            "type": "object",
            "properties": {
                "hostname": {"type": "string"},
                "port": {"type": "integer"},
                "username": {"type": "string"},
                "password": {"type": "string"}
            },
            "required": ["hostname", "username", "password"],
            "additionalProperties": False
        }
    },

    "type": "object",
    "properties": {
        "cache": {"$ref": "#/definitions/cache"},
        "auth": {
            "type": "array",
            "minItems": 1,
            "items": {"$ref": "#/definitions/datacenter"}
        }
    },
    "required": ["auth"],
    "additionalProperties": False
}


def parse_config_and_add_defaults(s):
    config = json.loads(s)
    jsonschema.validate(config, CONFIG_SCHEMA)
    for dc in config['auth']:
        dc.setdefault('port', CONFIG_DEFAULT_PORT)
    input_cache = config.get('cache', {})
    config['cache'] = copy.copy(CONFIG_DEFAULT_CACHE_PARAMS)
    config['cache'].update(input_cache)

    if config['cache']['expiration_seconds'] > 0:
        # verify cache file is writeable & is parseable if exists
        already_exists = os.path.exists(config['cache']['filename'])
        with open(config['cache']['filename'], 'a') as f:
            pass
        if not already_exists:
            os.unlink(config['cache']['filename'])
        else:
            with open(config['cache']['filename'], 'r') as f:
                json.loads(f.read())

    return config



import json
import jsonschema
import os
import tempfile
import time
import pytest
import vcenter_info
from vcenter_info import vcenter
from vcenter_info import api


VERSION_RESPONSE_SCHEMA = {
    '$schema': 'http://json-schema.org/draft-07/schema#',
    'type': 'object',
    'properties': {
        'api': {'type': 'string'},
        'module': {'type': 'string'}
    },
    'required': ['api', 'module'],
    'additionalProperties': False

}

VMLIST_SCHEMA = {
    '$schema': 'http://json-schema.org/draft-07/schema#',

    'definitions': {
        'stats': {
            'type': 'object',
            'properties': {
                'balloonedMemory': {'type': 'integer', 'minimum': -1},
                'compressedMemory': {'type': 'integer', 'minimum': -1},
                'consumedOverheadMemory': {'type': 'integer', 'minimum': -1},
                'distributedCpuEntitlement': {
                    'type': 'integer', 'minimum': -1},
                'distributedMemoryEntitlement': {
                    'type': 'integer', 'minimum': -1},
                'ftLatencyStatus': {'type': 'string'},
                'ftLogBandwidth': {'type': 'integer', 'minimum': -1},
                'ftSecondaryLatency': {'type': 'integer', 'minimum': -1},
                'guestHeartbeatStatus': {'type': 'string'},
                'guestMemoryUsage': {'type': 'integer', 'minimum': -1},
                'hostMemoryUsage': {'type': 'integer', 'minimum': -1},
                'overallCpuDemand': {'type': 'integer', 'minimum': -1},
                'overallCpuUsage': {'type': 'integer', 'minimum': -1},
                'privateMemory': {'type': 'integer', 'minimum': -1},
                'sharedMemory': {'type': 'integer', 'minimum': -1},
                'ssdSwappedMemory': {'type': 'integer', 'minimum': -1},
                'staticCpuEntitlement': {'type': 'integer', 'minimum': -1},
                'staticMemoryEntitlement': {'type': 'integer', 'minimum': -1},
                'swappedMemory': {'type': 'integer', 'minimum': -1},
                'uptimeSeconds': {'type': 'integer', 'minimum': -1}
            },
            'required': [
                'balloonedMemory',
                'compressedMemory',
                'consumedOverheadMemory',
                'distributedCpuEntitlement',
                'distributedMemoryEntitlement',
                'ftLatencyStatus',
                'ftLogBandwidth',
                'ftSecondaryLatency',
                'guestHeartbeatStatus',
                'guestMemoryUsage',
                'hostMemoryUsage',
                'overallCpuDemand',
                'overallCpuUsage',
                'privateMemory',
                'sharedMemory',
                'ssdSwappedMemory',
                'staticCpuEntitlement',
                'staticMemoryEntitlement',
                'swappedMemory',
                'uptimeSeconds'
            ],
            'additionalProperties': False
        },
        'vm': {
            'type': 'object',
            'properties': {
                'annotation': {'type': 'string'},
                'boot': {'type': ['string', 'null']},
                'datacenter': {'type': 'string'},
                'guest': {'type': 'string'},
                'ip': {'type': ['string', 'null']},
                'name': {'type': 'string'},
                'overallStatus': {'type': 'string'},
                'path': {'type': 'string'},
                'question': {'type': ['string', 'null']},
                'san': {'type': ['string', 'null']},
                'state': {'type': 'string'},
                'stats': {'$ref': '#/definitions/stats'}
            },
            'required': [
                'annotation', 'datacenter', 'guest',
                'ip', 'name', 'overallStatus', 'path',
                'question', 'san', 'state', 'stats'],
            'additionalProperties': False
        }
    },

    'type': 'array',
    'items': {'$ref': '#/definitions/vm'}

}


@pytest.fixture
def client(config_file):
    os.environ['CONFIG_FILENAME'] = config_file
    with vcenter_info.create_app().test_client() as c:
        yield c


@pytest.fixture
def dummy_json_file():
    with tempfile.NamedTemporaryFile() as f:
        f.write(json.dumps({'a': 1}).encode('utf-8'))
        f.flush()
        yield f.name


def test_vcenter_content_parsing(mocked_vcenter, auth_config):
    # don't fork for one test, for coverage
    values = vcenter.get_vms(auth_config, fork=False)
    jsonschema.validate(list(values), VMLIST_SCHEMA)


def test_version_api_response(client):
    rv = client.get('/api/version', headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    assert rv.is_json
    response_data = json.loads(rv.data.decode('utf-8'))
    jsonschema.validate(response_data, VERSION_RESPONSE_SCHEMA)


def test_vmlist_api_response(mocked_vcenter, client):
    rv = client.get('/api/vms', headers={'Accept': ['application/json']})
    assert rv.status_code == 200
    assert rv.is_json
    response_data = json.loads(rv.data.decode('utf-8'))
    jsonschema.validate(response_data, VMLIST_SCHEMA)


def test_cache_loading(dummy_json_file):
    vm_list = api.load_cached_vms({
        'filename': dummy_json_file,
        'expiration_seconds': 100
    })
    assert vm_list


def test_cache_expired(dummy_json_file):
    time.sleep(0.5)
    vm_list = api.load_cached_vms({
        'filename': dummy_json_file,
        'expiration_seconds': .001
    })
    assert vm_list is None


def test_cache_no_file():
    with tempfile.NamedTemporaryFile() as f:
        non_existent_filename = f.name
    vm_list = api.load_cached_vms({
        'filename': non_existent_filename,
        'expiration_seconds': 100
    })
    assert vm_list is None


def test_cache_bad_json():
    with tempfile.NamedTemporaryFile() as f:
        f.write('not json'.encode('utf-8'))
        f.flush()
        vm_list = api.load_cached_vms({
            'filename': f.name,
            'expiration_seconds': 100
        })
        assert vm_list is None



from vcenter_info import cli
from click.testing import CliRunner


def test_happy_flow(mocked_vcenter, config_file):
    expected_measurement = 'abab'
    runner = CliRunner()
    result = runner.invoke(
        cli.cli, [
            '--measurement', expected_measurement,
            '--config', config_file
        ])
    assert result.exit_code == 0
    lines = result.output.splitlines()
    assert lines  # there should be some output
    # TODO: verify influx format if necessary



import json
import logging
import re
import ssl
from multiprocessing import Process, Queue
from pyVim.connect import SmartConnect, Disconnect
from pyVmomi import vim
import datetime


logger = logging.getLogger(__name__)
MAX_DEPTH = 10


def vm_to_dict(vm):
    summary = vm.summary
    config = summary.config
    info = {
        'name': config.name,
        'path': config.vmPathName,
        'guest': config.guestFullName,
        'annotation': config.annotation,
        'state': str(summary.runtime.powerState),
        'overallStatus': str(summary.overallStatus),
        'ip': None,
        'question': None,
        'san': None,
        'boot': None,
        'stats': {
            # cf. https://vdc-download.vmware.com/vmwb-repository/dcr-public/da47f910-60ac-438b-8b9b-6122f4d14524/16b7274a-bf8b-4b4c-a05e-746f2aa93c8c/doc/vim.vm.Summary.QuickStats.html  # noqa
            'balloonedMemory': -1,
            'compressedMemory': -1,
            'consumedOverheadMemory': -1,
            'distributedCpuEntitlement': -1,
            'distributedMemoryEntitlement': -1,
            'ftLatencyStatus': None,
            'ftLogBandwidth': -1,
            'ftSecondaryLatency': -1,
            'guestHeartbeatStatus': None,
            'guestMemoryUsage': -1,
            'hostMemoryUsage': -1,
            'overallCpuDemand': -1,
            'overallCpuUsage': -1,
            'privateMemory': -1,
            'sharedMemory': -1,
            'ssdSwappedMemory': -1,
            'staticCpuEntitlement': -1,
            'staticMemoryEntitlement': -1,
            'swappedMemory': -1,
            'uptimeSeconds': -1
        }
    }
    if summary.guest is not None:
        info['ip'] = summary.guest.ipAddress
    if summary.runtime.question is not None:
        info['question'] = summary.runtime.question.text
    m = re.match(r'^.*\[([^]]+)\].*', info['path'])
    if m:
        info['san'] = m.group(1)

    if isinstance(summary.runtime.bootTime, datetime.datetime):
        info['boot'] = summary.runtime.bootTime.isoformat()

    for name in info['stats']:
        value = getattr(summary.quickStats, name, None)
        info['stats'][name] = value if isinstance(value, int) else str(value)

    return info


def vms_from_list(vmList, depth=0):
    if depth > MAX_DEPTH:
        return

    for vm in vmList:

        if hasattr(vm, 'childEntity'):
            yield from vms_from_list(vm.childEntity, depth+1)
            continue

        if isinstance(vm, vim.VirtualApp):
            yield from vms_from_list(vm.vm, depth+1)
            continue

        assert isinstance(vm, vim.VirtualMachine)
        yield vm


def load_vms_from_datacenter(server_auth_config):

    context = None
    if hasattr(ssl, '_create_unverified_context'):
        context = ssl._create_unverified_context()

    si = None
    try:
        port = server_auth_config.get('port', 443)
        si = SmartConnect(
            host=server_auth_config['hostname'],
            port=port,
            user=server_auth_config['username'],
            pwd=server_auth_config['password'],
            sslContext=context)

        if not si:
            print('error connecting to'
                  f' {server_auth_config["hostname"]}:{port}')
            return -1

        content = si.RetrieveContent()
        for datacenter in content.rootFolder.childEntity:
            if hasattr(datacenter, 'vmFolder'):
                yield {
                    'datacenter': datacenter.name,
                    'vms': vms_from_list(datacenter.vmFolder.childEntity)
                }

    finally:
        if si:
            Disconnect(si)


def _get_vms_from_server_proc(queue, server_auth_config):
    for dc in load_vms_from_datacenter(server_auth_config):
        for vm in dc['vms']:
            summary = vm_to_dict(vm)
            summary['datacenter'] = dc['datacenter']
            queue.put(summary)

    # contract: send None to indicate end of processing
    queue.put(None)


def _get_vms_no_fork(auth_config):
    """
    for test coverage only
    :param auth_config:
    :return:
    """
    q = Queue()
    for server in auth_config:
        _get_vms_from_server_proc(q, server)
        while True:
            vm = q.get()
            if vm:
                yield vm
            else:
                # contract: None indicates end of processing
                break


def get_vms(auth_config, fork=True):
    """
    create a child process for each server

    :param auth_config:
    :return:
    """

    if not fork:
        yield from _get_vms_no_fork(auth_config)
        return

    processes = []
    q = Queue()

    for server in auth_config:
        p = Process(
            target=_get_vms_from_server_proc,
            args=(q, server))
        p.start()
        processes.append(p)

    num_finished = 0
    while num_finished < len(processes):
        vm = q.get()
        if vm:
            yield vm
        else:
            # contract: None indicates end of processing
            num_finished += 1

    for p in processes:
        p.join()


if __name__ == "__main__":

    import jinja2
    # import jsonschema
    import os

    TEMPLATE_FILENAME = os.path.join(
        os.path.dirname(__file__),
        'report.html.tpl')
    CONFIG_FILENAME = 'config.json'

    with open(CONFIG_FILENAME) as f:
        config = json.loads(f.read())
        # jsonschema.validate(config,  CONFIG_SCHEMA)

    with open(TEMPLATE_FILENAME) as f:
        template = jinja2.Template(f.read())

    report = template.render(vms=get_vms(config))

    print(report)



"""
automatically invoked app factory
"""
import logging
import os

from flask import Flask

from vcenter_info import config

logger = logging.getLogger(__name__)
CONFIG_ENV_VAR_NAME = 'CONFIG_FILENAME'


def create_app():
    """
    :return: a new flask app instance
    """

    assert CONFIG_ENV_VAR_NAME in os.environ, \
        'expected {} to be defined in the environment'.format(
            CONFIG_ENV_VAR_NAME)
    config_filename = os.environ[CONFIG_ENV_VAR_NAME]
    assert os.path.isfile(config_filename), \
        'config file {} not found'.format(config_filename)

    with open(config_filename) as f:
        config_params = config.parse_config_and_add_defaults(f.read())

    app = Flask(__name__)
    app.secret_key = 'super secret session key'
    app.config['CONFIG_PARAMS'] = config_params

    from vcenter_info import api
    app.register_blueprint(api.blueprint, url_prefix='/api')

    from vcenter_info import default
    app.register_blueprint(default.blueprint, url_prefix='/')

    logger.info('vcenter_info Flask app initialized')

    return app



import functools
import json
import logging
import os
import pkg_resources
import time

from flask import Blueprint, jsonify, request, Response, current_app

from vcenter_info import vcenter

blueprint = Blueprint("vcenter-info-api-routes", __name__)

API_VERSION = '0.1'
RESPONSE_TIMEOUT_SEC = 2.0
logger = logging.getLogger(__name__)


def require_accepts_json(f):
    """
    used as a route handler decorator to return an error
    unless the request allows responses with type "application/json"
    :param f: the function to be decorated
    :return: the decorated function
    """
    @functools.wraps(f)
    def decorated_function(*args, **kwargs):
        # TODO: use best_match to disallow */* ...?
        if not request.accept_mimetypes.accept_json:
            return Response(
                response="response will be json",
                status=406,
                mimetype="text/html")
        return f(*args, **kwargs)
    return decorated_function


@blueprint.route("/version", methods=['GET', 'POST'])
@require_accepts_json
def version():
    version_params = {
        'api': API_VERSION,
        'module': pkg_resources.get_distribution('vcenter_info').version
    }
    return jsonify(version_params)


def load_cached_vms(params):
    if params['expiration_seconds'] > 0 \
            and os.path.isfile(params['filename']):
        stat = os.stat(params['filename'])
        if time.time() - stat.st_mtime < params['expiration_seconds']:
            with open(params['filename']) as f:
                try:
                    return json.loads(f.read())
                except json.JSONDecodeError:
                    logger.exception(f'error parsing {params["filename"]}')

    return None


@blueprint.route("/vms", methods=['GET', 'POST'])
@require_accepts_json
def vms():

    refresh = request.args.get('refresh', default=False, type=bool)

    logger.debug('getting vms')

    config = current_app.config['CONFIG_PARAMS']

    vm_list = None if refresh else load_cached_vms(config['cache'])
    if not vm_list:
        vm_list = list(vcenter.get_vms(config['auth']))
        with open(config['cache']['filename'], 'w') as f:
            f.write(json.dumps(vm_list))

    return jsonify(vm_list)



import logging
from flask import Blueprint, render_template\

blueprint = Blueprint("vcenter-info-default-routes", __name__)
logger = logging.getLogger(__name__)


@blueprint.route('/')
def index():
    return render_template('index.html')



import logging
import click
from vcenter_info import vcenter, config

logger = logging.getLogger(__name__)


def _key_values(d):
    return ['%s=%s' % (k, v) for k, v in d.items()]


def _make_influx_line(measurement, fields, tags):
    return '{measurement},{tags} {fields}'.format(
        measurement=measurement,
        tags=','.join(_key_values(tags)),
        fields=','.join(_key_values(fields)))


def _validate_config(ctx, param, value):
    """
    loads, validates and returns configuration parameters

    :param ctx:
    :param param:
    :param value: filename (string)
    :return: a dict containing configuration parameters
    """
    try:
        return config.parse_config_and_add_defaults(value.read())
    except Exception as e:
        raise click.BadParameter(e)


@click.command()
@click.option(
    '--measurement',
    required=True,
    type=click.STRING,
    help='influx measurement name')
@click.option(
    '--config',
    required=True,
    type=click.File('r'),
    help='config filename',
    callback=_validate_config)
def cli(measurement, config):

    for vm in vcenter.get_vms(config['auth']):
        if vm['state'] != 'poweredOn':
            continue
        tags = {'name': vm['name'], 'datacenter': vm['datacenter']}
        fields = dict([
            (k, v) for k, v in vm['stats'].items()
            if isinstance(v, int)])
        print(_make_influx_line(
            measurement=measurement,
            fields=fields,
            tags=tags))


if __name__ == '__main__':
    logging.basicConfig(level=logging.DEBUG)
    cli()



from setuptools import setup, find_packages

with open("README.md") as f:
    long_description = f.read()

setup(
    name='vcenter-info',
    version="0.0.4",
    author='Erik Reid',
    author_email='nobody@nowhere.org',
    description='Simple view of vCenter data',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/susrik/vcenter-info',
    packages=find_packages(),
    install_requires=[
        'pyvmomi',
        'jinja2',
        'jsonschema',
        'flask',
        'click'
    ],
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'check-vcenter-vms=vcenter_info.cli:cli'
        ]
    },
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Development Status :: 4 - Beta'
    ],
    python_requires='>=3.6'
)

